from .archiveit_collection import ArchiveItCollection, ArchiveItCollectionException
from .timemap import convert_LinkTimeMap_to_dict, MalformedLinkFormatTimeMap
from .archive_information import generate_raw_urim
from .utils import generate_archiveit_urits, process_timemaps_for_mementos, discover_raw_urims, get_uri_responses
from .version import name, version, user_agent_string


__all__ = [ "ArchiveItCollection", "ArchiveItCollectionException",
    "convert_LinkTimeMap_to_dict", "MalformedLinkFormatTimeMap",
    "generate_raw_urim", "generate_archiveit_urits", "process_timemaps_for_mementos",
    "discover_raw_urims", "get_uri_responses", "version", "name", "user_agent_string"
]

import logging
try:  # Python 2.7+
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

logging.getLogger(__name__).addHandler(NullHandler())


# -*- coding: utf-8 -*-

"""
aiu.archiveit_collection
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module acquires data about an Archvie-It collection both from its results
pages and its CSV seed report.

Note: There be dragons here. This code was originally written for a different 
project. I'm sure it can be improved.
"""

# TODO: allow one to specify a session object from requests

import os
import json
import requests
import logging
import logging.config
import argparse
import csv
import sys

from io import StringIO

from datetime import datetime
from bs4 import BeautifulSoup

from .version import user_agent_string

logger = logging.getLogger(__name__)

collection_uri_prefix = "https://archive-it.org/collections"

class ArchiveItCollectionException(Exception):
    """An exception class to be used by the functions in this file so that the
    source of error can be detected.
    """
    pass

def scrape_main_collection_data(soup):
    """Scrapes general collection metadata the Archive-It collection
    results page using the BeautfulSoup object `soup`.
    """

    data = {}

    metadata_tags = soup.find_all("div", "entity-meta")

    try:
        data["name"] = metadata_tags[0].find_all("h1")[0].get_text().strip()
        data["exists"] = True
    except IndexError:

        try:

            if "Not Found" in metadata_tags[0].find_all("h2")[0].get_text().strip():
                data["exists"] = False
                return data
            else:
                raise IndexError("Could not find 'Not Found' string using screen scraping")

        except IndexError as e:
            logger.error("Failed to find collection name using screen scraping")
            logger.error("Failed to determine if collection does not exist using screen scraping")
            logger.error(metadata_tags)
            raise e

    try:
        data["collection_uri"] = metadata_tags[0].find_all("h1")[0].find_all("a")[0]["href"]
    except IndexError as e:
        logger.error("Failed to find collection URI using screen scraping")
        logger.error(metadata_tags)
        raise e

    try:
        data["institution_name"] = \
            metadata_tags[0].find_all("p", "collectedBy")[0].get_text().strip().replace("Collected by:", "").strip()
    except IndexError as e:
        logger.error("Failed to find institution name using screen scraping")
        logger.error(metadata_tags)
        raise e

    try:
        data["institution_uri"] = metadata_tags[0].find_all("p", "collectedBy")[0].find_all("a")[0]['href']
    except IndexError as e:
        logger.error("Failed to find institution URI using screen scraping")
        logger.error(metadata_tags)
        raise e

    try:
        data["description"] = \
            metadata_tags[0].find_all("p", "seamTextPara")[0].get_text().strip().replace('\n', '<br />')
    except IndexError:
        logger.warning("Failed to find description using screen scraping, "
            "setting to No description...")
        data["description"] = "No description."

    search_results = soup.find_all("div", {"id": "all-search-results"})

    try:
        public_stmt = search_results[0].find_all("h2")[0].get_text().strip()

        if "There is no public content available for this Collection yet. Please check back soon!" in public_stmt:
            data["public"] = "private"
        else:
            data["public"] = "public"

    except IndexError:
        data["public"] = "unknown" 

    data["archived_since"] = "unknown"

    subjects = []

    for item in metadata_tags[0].find_all("p"):

        bolds = item.find_all('b')

        if len(bolds) > 0:

            for bold in bolds:

                if "Subject" in bold.get_text():
                    subjects = [i.strip(',').strip() for i in item.text.replace('Subject:', '').replace('\xa0', '').replace('\t', '').strip().split('\n')]

                if "Archived since" in bold.get_text():
                    archived_since = [i.strip(',').strip() for i in item.text.replace('Archived since:', '').replace('\xa0', '').replace('\t', '').strip().split('\n')][0]

    data["subject"] = subjects
    data["archived_since"] = archived_since

    return data

def scrape_optional_collection_data(soup):
    """Scrapes optional collection metadata the Archive-It collection
    results page using the BeautfulSoup object `soup`.
    """

    data = {}

    metadata_tags = soup.find_all("div", "entity-meta")

    if len(metadata_tags) > 0:

        moreMetadata = metadata_tags[0].find_all("div", "moreMetadata")

        if len(moreMetadata) > 0:

            for item in moreMetadata[0].find_all("p"):
                key = item.find_all("b")[0].text.replace('\xa0', '').strip().strip(':')
                values = [i.strip(',').strip() for i in item.text.replace("{}:".format(key), '').replace('\xa0', '').replace('\t', '').strip().split('\n')]
                key = key.lower()
                data[key] = values

    return data

def scrape_seed_metadata(soup):
    """Scrapes the seed metadata from an Archive-It results page stored in the
    BeautifulSoup object `soup`.
    """

    data = []

    for item in soup.find_all("div", "result-item"):

        itemdict = {}

        for entry in item.find_all("h3", "url"):
    
            if 'URL:' in entry.text:
                url = entry.text.replace("URL:", "").strip()
                itemdict["uri"] = url

            if 'Title:' in entry.text:
                title = entry.text.replace("Title:", "").strip()
                itemdict["title"] = title

        for entry in item.find_all("p"):
            if len(entry.find_all("b")) > 0:
                key = entry.find_all("b")[0].text.replace('\xa0', '').strip().rstrip(':')
                #value = entry.find_all("a")[0].text.strip()
                values = [i.strip(',').strip() for i in entry.text.replace("{}:".format(key), '').replace('\xa0', '').replace('\t', '').strip().split('\n')]
                itemdict[key.lower()] = values

        data.append(itemdict)

    return data

def scrape_page_count(soup):
    """Scrapes the page count from an Archive-It results page stored in the
    BeautifulSoup object `soup`.
    """

    try:
        pagestring = soup.find_all("div", "paginator")[0].text
        page_count = pagestring.split("of")[1].strip().split()[0]
        page_count = page_count.replace(',', '')
    except IndexError:
        page_count = None

    return page_count

def scrape_page_number(soup):
    """Scrapes the page number from an Archive-It results page stored in the
    BeautifulSoup object `soup`.
    """

    try:
        pagestring = soup.find_all("div", "paginator")[0].text
        print("pagestring: {}".format(pagestring))
        page_number = pagestring.split("of")[0].strip().split()[1]
        page_number = page_number.replace(',', '')
    except IndexError:
        page_number = None

    return page_number

def scrape_result_count(soup):
    """Scrapes the result count from an Archive-It results page stored in the
    BeautifulSoup object `soup`.
    """

    try:
        pagestring = soup.find_all("div", "paginator")[0].text
        result_count = pagestring[pagestring.find('(') + 1:].replace(' Total Results', '').rstrip(')')
        result_count = pagestring[pagestring.find('(') + 1:pagestring.find(' Total Results')]
        result_count = result_count.replace(',', '')
    except IndexError:
        result_count = None

    return result_count


def scrape_next_page_number(soup):
    """Scrapes the Next Page URI from an Archive-It results page stored in
    the BeautifulSoup object `soup`.
    """

    nextpage = None

    nextpagematch = soup.find_all('a', {'id': 'pageNext'})

    if len(nextpagematch) == 2:
        nextpage_uri = nextpagematch[0]['href'] 
        nextpage = nextpage_uri[nextpage_uri.find('page='):][:nextpage_uri.find('&') - 1].replace("page=", "") 

    return nextpage

def get_seed_metadata_from_seed_report(collection_id, session):
    """Builds the CSV seed report URI using `collection_id` and saves the
    seed report in `pages_dir`.
    """

    seed_metadata = {}

    seed_report_uri = "https://partner.archive-it.org/api/seed?" \
        "annotate__max=seed_group__name&annotate__max=crawl_definition__recurrence_type" \
        "&collection={}&csv_header=Seed+URL&csv_header=Group&csv_header=Status" \
        "&csv_header=Frequency&csv_header=Type&csv_header=Access&format=csv&" \
        "show_field=url&show_field=seed_group__name&show_field=active&" \
        "show_field=crawl_definition__recurrence_type&show_field=seed_type&" \
        "show_field=publicly_visible&sort=-id".format(collection_id)

    r = session.get(seed_report_uri, headers={'user-agent': user_agent_string})

    seed_report = StringIO(r.text)
    
    csvreader = csv.DictReader(seed_report, delimiter=',')

    for row in csvreader:
    
        seed_metadata[ row["Seed URL"] ] = {} 
        seed_metadata[ row["Seed URL"] ]["group"] = row["Group"]
        seed_metadata[ row["Seed URL"] ]["status"] = row["Status"]
        seed_metadata[ row["Seed URL"] ]["frequency"] = row["Frequency"]
        seed_metadata[ row["Seed URL"] ]["type"] = row["Type"]
        seed_metadata[ row["Seed URL"] ]["access"] = row["Access"]

    return seed_metadata 

class ArchiveItCollection:
    """Organizes all information acquired about the Archive-It collection."""

    def __init__(self, collection_id, session=requests.Session(), logger=None):

        self.collection_id = str(collection_id)
        self.session = session
        self.metadata_loaded = False
        self.seed_metadata_loaded = False
        self.metadata = {}
        self.seed_metadata = {}
        self.private = None
        self.exists = None

        self.collection_uri = "{}/{}".format(collection_uri_prefix, collection_id)
        self.firstpage_response = self.session.get(self.collection_uri)        

        self.logger = logger or logging.getLogger(__name__)

    def load_collection_metadata(self):
        """Loads collection metadata from an existing directory, if possible.
        If the existing directory does not exist, it will then download the
        collection and process its metadata.
        """

        if not self.metadata_loaded:

            soup = BeautifulSoup(self.firstpage_response.text, 'html5lib')

            self.metadata["main"] = scrape_main_collection_data(soup)
            self.metadata["optional"] = scrape_optional_collection_data(soup)

            self.metadata_loaded = True

    def load_seed_metadata(self):
        """Loads the seed metadata previously downloaded if possible, otherwise
        acquires all collection results pages and parses them for seed metadata.

        This function is separate to limit the number of requests. It should only
        be called if seed metadata is needed.
        """

        self.load_collection_metadata()

        if self.is_private() is True:
            self.seed_metadata = {}
            return

        if not self.seed_metadata_loaded:

            timestamp = datetime.now()

            seed_metadata_list = []

            soup = BeautifulSoup(self.firstpage_response.text, 'html5lib')

            seed_metadata_list.extend( scrape_seed_metadata(soup) )

            nextpage = scrape_next_page_number(soup)
            result_count = scrape_result_count(soup)

            while nextpage:

                page_uri = "{}/{}/?page={}&totalResultCount={}".format(
                    collection_uri_prefix, self.collection_id, 
                    nextpage, result_count)

                r = self.session.get(page_uri)

                soup = BeautifulSoup(r.text, 'html5lib')

                seed_metadata_list.extend( scrape_seed_metadata(soup) )

                nextpage = scrape_next_page_number(soup)

            for item in seed_metadata_list:
                uri = item["uri"]

                itemdict = {}

                for key in item:
                    if key != "uri":
                        itemdict[key] = item[key]

                self.seed_metadata.setdefault("seeds", {})
                self.seed_metadata["seeds"].setdefault(uri, {})
                self.seed_metadata["seeds"][uri].setdefault(
                    "collection_web_pages", []).append(itemdict)

            seed_report_metadata = get_seed_metadata_from_seed_report(self.collection_id, self.session)
            seed_report_timestamp = datetime.now()

            for uri in seed_report_metadata:
                self.seed_metadata.setdefault("seeds", {})
                self.seed_metadata["seeds"].setdefault(uri, {})
                self.seed_metadata["seeds"][uri].setdefault("seed_report", {})
                self.seed_metadata["seeds"][uri]["seed_report"]= seed_report_metadata[uri]

            self.seed_metadata["timestamps"] = {}

            self.seed_metadata["timestamps"]["seed_metadata_timestamp"] = timestamp

            self.seed_metadata["timestamps"]["seed_report_timestamp"] = seed_report_timestamp

            self.seed_metadata_loaded = True

    def get_collection_name(self):
        """Getter for the collection name, as scraped."""

        self.load_collection_metadata()

        name = self.metadata["main"]["name"]

        return name

    def get_collection_uri(self):
        """Getter for the collection URI, as constructed."""

        self.load_collection_metadata()

        uri = "https://archive-it.org{}".format(self.metadata["main"]["collection_uri"])

        return uri

    def get_collectedby_uri(self):
        """Getter for the collecting organization's URI, as constructed."""

        self.load_collection_metadata()

        uri = "https://archive-it.org{}".format(self.metadata["main"]["institution_uri"])

        return uri

    def get_description(self):
        """Getter for the collection description, as scraped."""

        self.load_collection_metadata()

        description = self.metadata["main"]["description"]

        return description

    def get_collectedby(self):
        """Getter for the collecting organization's name, as scraped."""

        self.load_collection_metadata()

        collectedby = self.metadata["main"]["institution_name"]

        return collectedby

    def get_subject(self):
        """Getter for the collection's topics, as scraped."""
        
        self.load_collection_metadata()

        subjects = self.metadata["main"]["subject"]

        return subjects

    def get_archived_since(self):
        """Getter for the collection's archived since field, as scraped."""

        self.load_collection_metadata()

        archived_since = self.metadata["main"]["archived_since"]

        return archived_since

    def get_optional_metadata(self, key):
        """Given optional metadata field `key`, returns value as scraped."""

        self.load_collection_metadata()

        value = self.metadata["optional"][key]

        return value

    def list_optional_metadata_fields(self):
        """Lists the optional metadata fields, as scraped."""

        self.load_collection_metadata()

        keylist = self.metadata["optional"].keys()

        return keylist

    def is_private(self):
        """Returns `True` if the collection is public, otherwise `False`."""

        self.load_collection_metadata()

        if self.metadata["main"]["public"] == "private":
            return True
        elif self.metadata["main"]["public"] == "public":
            return False
        else:
            raise ArchiveItCollectionException("Could not determine private/public status")

    def does_exist(self):
        """Returns `True` if the collection actually exists and requests for
        its data did not result in 404s or soft-404s."""

        self.load_collection_metadata()

        exists = self.metadata["main"]["exists"]

        #self.logger.debug("exists is set to {}".format(exists))

        return exists

    def list_seed_uris(self):
        """Lists the seed URIs of an Archive-It collection."""

        self.load_collection_metadata()
        self.load_seed_metadata()

        return list(self.seed_metadata["seeds"].keys())

    def get_seed_metadata(self, uri):
        """Returns a `dict` of seed metadata associated with the memento 
        at `uri`."""

        d = self.seed_metadata["seeds"][uri] 

        return d

    def return_collection_metadata_dict(self):
        """Returns a `dict` of collection metadata."""

        self.load_collection_metadata()

        metadata = {
            "id": self.collection_id
        }    

        metadata["exists"] = self.does_exist()
        metadata["metadata_timestamp"] = \
            datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        if metadata["exists"]:

            metadata["name"] = self.get_collection_name()
            metadata["uri"] = self.get_collection_uri()
            metadata["collected_by"] = self.get_collectedby()
            metadata["collected_by_uri"] = self.get_collectedby_uri()
            metadata["description"] = self.get_description()
            metadata["subject"] = self.get_subject()
            metadata["archived_since"] = self.get_archived_since()
            metadata["private"] = self.is_private()
            metadata["optional"] = {}

            for key in self.list_optional_metadata_fields():
                metadata["optional"][key] = self.get_optional_metadata(key)

        return metadata

    def return_seed_metadata_dict(self):
        """Returns all seed metadata of all mementos as a `dict`."""

        self.load_seed_metadata()

        metadata = {
            "id": self.collection_id
        }

        metadata["seed_metadata"] = self.seed_metadata

        if len(metadata["seed_metadata"]) > 0:

            metadata["seed_metadata"]["timestamps"]["seed_metadata_timestamp"] = \
                    self.seed_metadata["timestamps"]["seed_metadata_timestamp"].strftime("%Y-%m-%d %H:%M:%S")
    
            metadata["seed_metadata"]["timestamps"]["seed_report_timestamp"] = \
                    self.seed_metadata["timestamps"]["seed_report_timestamp"].strftime("%Y-%m-%d %H:%M:%S")

        return metadata


    def return_all_metadata_dict(self):
        """Returns all metadata of the collection as a `dict`."""

        self.load_collection_metadata()
        self.load_seed_metadata()

        collection_metadata = self.return_collection_metadata_dict()
        
        collection_metadata["seed_metadata"] = self.seed_metadata

        if len(collection_metadata["seed_metadata"]) > 0:

            collection_metadata["seed_metadata"]["timestamps"]["seed_metadata_timestamp"] = \
                self.seed_metadata["timestamps"]["seed_metadata_timestamp"].strftime("%Y-%m-%d %H:%M:%S")
    
            collection_metadata["seed_metadata"]["timestamps"]["seed_report_timestamp"] = \
                self.seed_metadata["timestamps"]["seed_report_timestamp"].strftime("%Y-%m-%d %H:%M:%S")

        return collection_metadata
        
    def save_all_metadata_to_file(self, filename):
        """Saves all metadata to `filename` in JSON format."""

        collection_metadata = self.return_all_metadata_dict()

        with open(filename, 'w') as metadata_file:
            json.dump(collection_metadata, metadata_file, indent=4)




# -*- coding: utf-8 -*-

"""
aiu.utils
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module contains functions used to download mementos and TimeMaps and 
process them.

Note: There be dragons here. This code was originally written for a different 
project. I'm sure it can be improved.
"""

import os
import logging
import multiprocessing
import csv
import json
import hashlib
import random

from requests_futures.sessions import FuturesSession
from requests.exceptions import ConnectionError, TooManyRedirects

from .archive_information import generate_raw_urim
from .timemap import convert_LinkTimeMap_to_dict
from .version import user_agent_string

logger = logging.getLogger(__name__)
cpu_count = multiprocessing.cpu_count()

def get_head_responses(session, uris):
    """This function creates a futures object for each URI-M in `uris,
    using an existing `session` object from requests-futures. Only HEAD
    requests are performed.
    """ 

    futures = {}

    for uri in uris:

        logger.debug("issuing HEAD on uri {}".format(uri))

        futures[uri] = session.head(uri, 
            headers={'user-agent': user_agent_string},
            allow_redirects=True)

    return futures

def get_uri_responses(session, raw_uris):
    """This function creates a futures object for each URI-M in `raw_uris`,
    using an existing `session` object from requests-futures. Only GET
    requests are performed.
    """

    futures = {}

    for uri in raw_uris:

        logger.debug("issuing GET on uri {}".format(uri))

        futures[uri] = session.get(uri, 
            headers={'user-agent': user_agent_string},
            stream=True)

    return futures

def generate_archiveit_urits(cid, seed_uris):
    """This function generates TimeMap URIs (URI-Ts) for a list of `seed_uris`
    from an Archive-It colleciton specified by `cid`.
    """

    urit_list = []

    for urir in seed_uris:
        urit = "http://wayback.archive-it.org/{}/timemap/link/{}".format(
            cid, urir
        )

        urit_list.append(urit)  

    return urit_list

def discover_raw_urims(urimlist, futures=None):
    """This function checks that the URI-Ms in `urimlist` are valid mementos,
    following all redirects and checking for a Memento-Datetime header.
    """

    raw_urimdata = {}
    errordata = {}

    if futures == None:
        with FuturesSession(max_workers=cpu_count) as session:
            futures = get_head_responses(session, urimlist)

    working_uri_list = list(futures.keys())

    completed_urims = []

    # for urim in list_generator(working_uri_list):
    while len(completed_urims) < len(list(working_uri_list)):

        urim = random.choice(
            list(set(working_uri_list) - set(completed_urims))
        )

        logger.debug("checking if URI-M {} is done downloading".format(urim))

        if futures[urim].done():

            logger.debug("searching for raw version of URI-M {}".format(urim))

            try:

                response = futures[urim].result()

                if "memento-datetime" in response.headers:

                    if len(response.history) == 0:
                        raw_urimdata[urim] = generate_raw_urim(urim)
                    else:
                        raw_urimdata[urim] = generate_raw_urim(response.url)

                    logger.debug("added raw URI-M {} associated with URI-M {}"
                        " to the list to be downloaded".format(raw_urimdata[urim], urim))

                else:

                    warn_msg = "No Memento-Datetime in Response Headers for " \
                        "URI-M {}".format(urim)

                    logger.warning(warn_msg)
                    errordata[urim] = warn_msg

            except ConnectionError as e:
                logger.warning("While acquiring memento at {} there was an error of {}, "
                    "this event is being recorded".format(urim, repr(e)))
                errordata[urim] = repr(e)

            except TooManyRedirects as e:
                logger.warning("While acquiring memento at {} there was an error of {},"
                    "this event is being recorded".format(urim, repr(e)))
                errordata[urim] = repr(e)

            finally:
                logger.debug("Removing URI-M {} from the processing list".format(urim))
                completed_urims.append(urim)

    return raw_urimdata, errordata

def list_generator(input_list):
    """This function generates the next item in a list. It is useful for lists
    that have their items deleted while one is iterating through them.
    """

    logger.debug("list generator called")

    while len(input_list) > 0:
        for item in input_list:
            logger.debug("list now has {} items".format(len(input_list)))
            logger.debug("yielding {}".format(item))
            yield item

def process_timemaps_for_mementos(urit_list, working_directory):
    """This function acquires a list of mementos from a list of TimeMaps URIs.
    The TimeMaps are stored in `working_directory`.
    """

    timemap_data = {}

    output_directory = "{}/capture/timemaps".format(working_directory)

    if not os.path.isdir(output_directory):
        os.makedirs(output_directory)

    with FuturesSession(max_workers=cpu_count) as session:
        futures = get_uri_responses(session, urit_list)

    working_uri_list = list(futures.keys())

    with open("{}/manifest.tsv".format(output_directory), 'w') as manifestout:

        with open("{}/errors.jsonl".format(output_directory), 'w') as errorsout:

            fieldnames = [ "URI-T", "Filename"]

            manifestwriter = csv.DictWriter(manifestout, fieldnames, delimiter='\t')
            manifestwriter.writeheader()

            for urit in list_generator(working_uri_list):

                logger.debug("checking if URI-T {} is done downloading".format(urit))

                if futures[urit].done():

                    logger.debug("URI-T {} is done, extracting content".format(urit))

                    try:
                        response = futures[urit].result()

                        http_status = response.status_code

                        if http_status == 200:

                            timemap_content = response.text

                            logger.info("adding TimeMap content for URI-T {}".format(
                                urit))

                            uritfilename = hashlib.sha256(urit.encode('utf8')).hexdigest()

                            with open("{}/{}".format(
                                output_directory, uritfilename), 'w') as tmout:
                                tmout.write(timemap_content)

                            timemap_data[urit] = convert_LinkTimeMap_to_dict(
                                timemap_content, skipErrors=True)

                            manifestwriter.writerow({
                                'URI-T': urit,
                                'Filename': uritfilename
                            })

                        else:
                            errorsout.write("{}\n".format(json.dumps(
                                {
                                    "URI-T": urit,
                                    "status": http_status,
                                    "headers": dict(response.headers)
                                }
                            )))

                        # TODO: else store connection errors
                        working_uri_list.remove(urit)
                    
                    except ConnectionError as e:

                        logger.warning("There was a connection error while attempting "
                            "to download URI-T {}".format(urit))

                        errorsout.write(json.dumps(
                            {
                                "URI-T": urit,
                                "error": repr(e)
                            }
                        ))

                        # TODO: store connection errors in CollectionModel
                        working_uri_list.remove(urit)

                    except TooManyRedirects as e:

                        logger.warning("There were too many redirects while attempting "
                            "to download URI-T {}".format(urit))

                        errorsout.write(json.dumps(
                            {
                                "URI-T": urit,
                                "error": repr(e)
                            }
                        ))

                        # TODO: store connection errors in CollectionModel
                        working_uri_list.remove(urit)

    return timemap_data




# -*- coding: utf-8 -*-

"""
aiu.archive_information
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module handles the idiosyncracies between archives.
"""

archive_mappings = {
    "wayback.archive-it.org": ( '/http', 'id_/http' ),
    "web.archive.org": ( '/http', 'id_/http' )
}

def generate_raw_urim(urim):
    """Generates a raw URI-M based on the archive it belongs to. Supported
    URI patterns are found in `archive_mappings`.
    """

    raw_urim = urim

    for domainname in archive_mappings:

        if domainname in urim:

            search_pattern = archive_mappings[domainname][0]
            replacement_pattern = archive_mappings[domainname][1]

            # if urim is already a raw urim, do nothing
            if replacement_pattern not in urim:

                raw_urim = urim.replace(
                    search_pattern, replacement_pattern
                )

            break

    return raw_urim


import os
import codecs
import re

from setuptools import setup
from setuptools.command.install import install as _install
from os import path

# Python packaging info: http://python-packaging.readthedocs.io/en/latest/index.html
# More Python packaging info: http://python-packaging-user-guide.readthedocs.io/tutorials/distributing-packages/
# Python version info: https://www.python.org/dev/peps/pep-0440/

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

here = os.path.abspath(os.path.dirname(__file__))

# taken from https://packaging.python.org/guides/single-sourcing-package-version/
def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()

def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

setup(
    name='aiu',
    # cmdclass={'install': Install},
    version=find_version("aiu", "version.py"),
    description='Tools for for interacting with Archive-It.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/shawnmjones/archiveit_utilities',
    author='Shawn M. Jones',
    author_email='jones.shawn.m@gmail.com',
    license='MIT',
    packages=['aiu'],
    scripts=['bin/seeds2warc', 'bin/tm2warc', 'bin/fetch_ait_metadata'],
    install_requires=[
        'requests_futures',
        'warcio',
        'requests',
        'bs4',
        'html5lib',
        'requests_cache'
    ],
    # setup_requires=['nltk'],
    test_suite="tests",
    zip_safe=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='webarchives memento'
    )


# -*- coding: utf-8 -*-

"""
otmt.timemap
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module allows one to parse a link-format TimeMap.
"""

from datetime import datetime

class MalformedLinkFormatTimeMap(Exception):
    """
        This class exists to indicate errors while processing TimeMaps in
        link format.
    """
    pass

def convert_LinkTimeMap_to_dict(timemap_text, skipErrors=False):
    """
        A function to convert the link format TimeMap text into a Python 
        dictionary that closely resembles the JSON specified at:
        http://mementoweb.org/guide/timemap-json/

        There is one difference: the value of the datetime attribute is
        an actual Python datetime object.

        One can set skipErrors to True in order to skip errors in processing
        the TimeMap, but use with caution as it can lead to unpredictable
        behavior.
    """

    def process_local_dict(local_dict, working_dict):

        first = False
        last = False

        for uri in local_dict:

            relation = local_dict[uri]["rel"]

            local_memento_dict = {}

            if relation == "original":
                working_dict["original_uri"] = uri

            elif relation == "timegate":
                working_dict["timegate_uri"] = uri

            elif relation == "self":
                working_dict["timemap_uri"] = {}
                working_dict["timemap_uri"]["link_format"] = uri

            elif "memento" in relation:
                working_dict.setdefault("mementos", {})

                if "first" in relation:
                    working_dict["mementos"]["first"] = {}
                    working_dict["mementos"]["first"]["uri"] = uri
                    first = True

                if "last" in relation:
                    working_dict["mementos"]["last"] = {}
                    working_dict["mementos"]["last"]["uri"] = uri
                    last = True

                working_dict["mementos"].setdefault("list", [])

                local_memento_dict["datetime"] = None
                local_memento_dict["uri"] = uri

            if "datetime" in local_dict[uri] and "memento" in relation:

                mdt = datetime.strptime(local_dict[uri]["datetime"],
                    "%a, %d %b %Y %H:%M:%S GMT")

                local_memento_dict["datetime"] = mdt

                working_dict["mementos"]["list"].append(local_memento_dict)

                if first:
                    working_dict["mementos"]["first"]["datetime"] = mdt

                if last:
                    working_dict["mementos"]["last"]["datetime"] = mdt
                
        return working_dict


    dict_timemap = {}

    # current_char = ""
    uri = ""
    key = ""
    value = ""
    local_dict = {}
    state = 0
    charcount = 0

    for character in timemap_text:
        charcount += 1

        if state == 0:

            local_dict = {}
            uri = ""

            if character == '<':
                state = 1
            elif character.isspace():
                pass
            else:
                if not skipErrors:
                    raise MalformedLinkFormatTimeMap(
                        "issue at character {} while looking for next URI"
                        .format(charcount))

        elif state == 1:

            if character == '>':
                # URI should be saved by this point
                state = 2
                uri = uri.strip()
                local_dict[uri] = {}
            else:
                uri += character

        elif state == 2:

            if character == ';':
                state = 3

            elif character.isspace():
                pass

            else:
                if not skipErrors:
                    raise MalformedLinkFormatTimeMap(
                        "issue at character {} while looking for relation"
                        .format(charcount))

        elif state == 3:

            if character == '=':
                state = 4
            else:
                key += character

        elif state == 4:

            if character == ';':
                state = 3
            elif character == ',':
                state = 0

                process_local_dict(local_dict, dict_timemap)

            elif character == '"':
                state = 5
            elif character.isspace():
                pass
            else:
                if not skipErrors:
                    raise MalformedLinkFormatTimeMap(
                        "issue at character {} while looking for value"
                        .format(charcount))

        elif state == 5:

            if character == '"':
                state = 4

                key = key.strip()
                value = value.strip()
                local_dict[uri][key] = value
                key = ""
                value = ""

            else:
                value += character

        else:
            
            if not skipErrors:
                raise MalformedLinkFormatTimeMap(
                    "discovered unknown state while processing TimeMap")

    process_local_dict(local_dict, dict_timemap)

    return dict_timemap